from __future__ import annotations

import os
from itertools import combinations
from multiprocessing import cpu_count
from pathlib import Path

import numpy as np
import pandas as pd
import structlog
from scipy.special import factorial
from textdistance import hamming

from hilary.expectmax import EM
from hilary.utils import applyParallel, create_classes, preprocess

pd.set_option("mode.chained_assignment", None)

log = structlog.get_logger(__name__)


class Apriori:
    """Computes statistics of pairwise distances"""

    def __init__(
        self,
        df: pd.DataFrame,
        lengths=np.arange(15, 81 + 3, 3).astype(int),
        nmax: int = int(1e5),
        precision: float = 0.99,
        sensitivity: float = 0.9,
        threads: int = cpu_count() - 1,
        model: int = 326713,
        silent: bool = False,
    ):
        """Initialize attributes to later run class methods.

        Args:
            df (pd.DataFrame): Dataframe of sequences
            lengths (_type_, optional): CDR3 lengths used to filter non productive sequences. \
                Defaults to np.arange(15, 81 + 3, 3).
            nmax (int, optional): For parameter inference, sample and use nmax sequences for \
                classes larger than nmax. Defaults to 100000.
            precision (float, optional): Desired precision, defaults to 0.99.
            sensitivity (float, optional): Desired sensitivity, defaults to 0.9.
            threads (int, optional): Number of cpus on which to run code, defaults to cpu_count().
            model (int, optional): Model name to infer Null distribution, defaults to 326713.
            silent (bool) : If true do not to show progress bars.
        """
        self.lengths = lengths
        self.nmax = nmax
        self.threads = threads
        self.df = df
        self.threads = threads
        self.precision = precision
        self.sensitivity = sensitivity
        self.silent = silent
        self.histograms = None
        self.mean_prevalence = None
        self.mean_mean_distance = None
        self.cdfs = pd.read_csv(
            Path(os.path.dirname(__file__)) / Path(f"cdfs_{model}.csv"),
        )
        self.preprocess()
        self.classes = self.create_classes()

    def preprocess(self) -> pd.DataFrame:
        """Remove non productive sequences from dataframe.

        Returns:
            pd.Dataframe: Dataframe self.df containing all sequences.
        """
        self.df = preprocess(
            self.df,
            silent=self.silent,
        )
        return self.df

    def create_classes(self) -> pd.DataFrame:
        """Create VJl and l classes from self.df."""
        self.classes = create_classes(self.df)
        return self.classes

    def vjls2x(self, args: tuple[int, pd.DataFrame]) -> pd.DataFrame:
        """Compute histogram for a given VJl class

        Returns:
            pd.DataFrame: Histogram of distances for a given large VJl class."""
        i, df = args
        xs = []
        for s1, s2 in combinations(df["cdr3"].values, 2):
            xs.append(hamming(s1, s2))
        return pd.DataFrame(
            np.histogram(
                xs,
                bins=range(
                    self.lengths[-1] + 2,
                ),
                density=False,
            )[0],
            columns=[i],
        ).transpose()

    def compute_allvjl(
        self,
        lengths: np.ndarray = np.arange(15, 81 + 3, 3).astype(int),
    ) -> pd.DataFrame:
        """Compute histograms for all large VJl classes.

        Returns:
            pd.DataFrame: Histogram of distances for large VJl classes."""
        query = "v_gene != 'None' and pair_count >0 and cdr3_length in @lengths"
        groups = self.df.groupby(["v_gene", "j_gene", "cdr3_length"])
        log.debug(
            "Computing CDR3 hamming distances within all large VJl classes.",
        )
        results = applyParallel(
            [
                (
                    row.class_id,
                    groups.get_group((row.v_gene, row.j_gene, row.cdr3_length)).sample(
                        frac=min(np.sqrt(self.nmax / row.pair_count), 1),
                    ),
                )
                for _, row in self.classes.query(query).iterrows()
            ],
            self.vjls2x,
            cpuCount=self.threads,
            silent=self.silent,
        )
        if results.empty:
            results = pd.DataFrame(columns=[*range(81 + 1)])
        results["class_id"] = results.index
        return results

    def get_histograms(self) -> pd.DataFrame:
        """Compute histograms for all large classes

        Returns:
            pd.DataFrame: Histogram of distances for all large classes."""
        hs_vjl = self.compute_allvjl()
        self.histograms = hs_vjl.sort_values(
            "class_id",
        )[["class_id"] + [*range(81 + 1)]]
        return self.histograms

    def estimate(self, args: tuple[tuple[int], pd.DataFrame]) -> pd.DataFrame:
        """Fit prevalence and mu using the histogram which is the distribution of distances.

        Args:
            args tuple[int, pd.DataFrame]: class_id, histogram for that class

        Returns:
            pd.DataFrame: dataframe with parameters for each class id.
        """
        class_id, h = args
        l = self.classes.loc[self.classes.class_id == class_id].cdr3_length.values[0]
        em = EM(l, h.values[0, 1:], positives="geometric")
        rho_geo, mu_geo = em.discreteEM()
        error_geo = em.error([rho_geo, mu_geo])
        em = EM(l, h.values[0, 1:], positives="poisson")
        rho_poisson, mu_poisson = em.discreteEM()
        error_poisson = em.error([rho_poisson, mu_poisson])
        result = pd.DataFrame(
            columns=[
                "class_id",
                "rho_geo",
                "mu_geo",
                "error_geo",
                "rho_poisson",
                "mu_poisson",
                "error_poisson",
            ],
        )
        result.class_id = [class_id[0]]
        result.rho_geo = [rho_geo]
        result.mu_geo = [mu_geo]
        result.error_geo = [error_geo]
        result.rho_poisson = [rho_poisson]
        result.mu_poisson = [mu_poisson]
        result.error_poisson = [error_poisson]
        return result

    def get_parameters(self) -> None:
        """Computes prevalence and mean distance for all classes."""
        if self.histograms is None:
            raise ValueError(
                "Histogram attribute is None. Please run get_histograms method.",
            )
        parameters = applyParallel(
            self.histograms.groupby(["class_id"]),
            self.estimate,
            cpuCount=self.threads,
            silent=self.silent,
        ).reset_index(drop=True)

        self.classes.index = self.classes.class_id
        parameters.index = parameters.class_id

        self.classes["prevalence"] = parameters[
            [
                "rho_poisson",
                "rho_geo",
            ]
        ].min(axis=1)
        self.classes["mean_distance"] = (
            parameters["mu_poisson"] / self.classes["cdr3_length"]
        )
        self.classes["effective_prevalence"] = self.classes["prevalence"].fillna(0.2)
        self.classes["effective_mean_distance"] = self.classes["mean_distance"].fillna(
            0.2,
        )

    def assign_precise_sensitive_thresholds(
        self,
        args: tuple[tuple[int], pd.DataFrame],
    ) -> pd.DataFrame:
        """Assign precise and sensitive thresholds to dataframe grouped by length.

        Args:
            args tuple[tuple[int],pd.DataFrame]: length, dataframe grouped by length.

        Returns:
            pd.DataFrame: Precise and sensitive thresholds.
        """

        tuple_l, ldf = args
        l = tuple_l[0]
        if l > 81 or l < 15 or l % 3 != 0:
            ldf[["precise_threshold", "sensitive_threshold"]] = (
                np.ones((len(ldf), 2), dtype=int) * l // 5
            )
            return ldf[["precise_threshold", "sensitive_threshold"]]

        rhos = ldf["effective_prevalence"]
        mus = ldf["effective_mean_distance"] * l
        bins = np.arange(l + 1)
        cdf0 = self.cdfs.loc[self.cdfs["l"] == l].values[0, 1 : l + 2]
        cdf1 = (
            np.array([mu**bins * np.exp(-mu) for mu in mus]) / factorial(bins)
        ).cumsum(axis=1)
        ps = cdf0 / cdf1
        t_sens = (cdf1 < self.sensitivity).sum(axis=1)
        t_prec = (
            np.array(
                [
                    p < rho / (1 + 1e-5 - rho) * (1 - self.precision) / self.precision
                    for p, rho in zip(ps, rhos)
                ],
            ).sum(axis=1)
            - 1
        )
        t_prec = np.min([t_prec, t_sens], axis=0)
        ldf["precise_threshold"] = t_prec
        ldf["sensitive_threshold"] = t_sens
        return ldf[["precise_threshold", "sensitive_threshold"]]

    def get_thresholds(self) -> None:
        """Assigns thresholds using null distribution from model.

        Args:
            model int: _description_. Defaults to 326713.
        """
        self.classes[["precise_threshold", "sensitive_threshold"]] = applyParallel(
            self.classes.groupby(["cdr3_length"]),
            self.assign_precise_sensitive_thresholds,
            cpuCount=self.threads,
            silent=self.silent,
        )



from __future__ import annotations

import logging
from multiprocessing import cpu_count
from pathlib import Path

import structlog
import typer
from scipy.special import binom

from hilary.apriori import Apriori
from hilary.inference import HILARy
from hilary.utils import pairwise_evaluation, read_input, save_dataframe

app = typer.Typer(add_completion=False)


@app.command()
def main(
    data_path: Path = typer.Argument(
        ...,
        help="Path of the excel file to infer lineages.",
        show_default=False,
    ),
    verbose: int = typer.Option(
        0,
        "--verbose",
        "-v",
        count=True,
        help="Set logging verbosity level.",
    ),
    threads: int = typer.Option(
        cpu_count(),
        "--threads",
        "-t",
        help="Choose number of cpus on which to run code.",
    ),
    precision: float = typer.Option(
        0.99,
        "--precision",
        "-p",
        help="Choose desired precision.",
    ),
    sensitivity: float = typer.Option(
        0.95,
        "--sensitivity",
        "-s",
        help="Choose desired sensitivity.",
    ),
    model: int = typer.Option(
        326713,
        "--model",
        "-m",
        help="Model name to infer Null distribution.",
    ),
    silent: bool = typer.Option(
        False,
        "--silent",
        help="Do not show progress bars if used.",
    ),
    result_folder: Path = typer.Option(
        None,
        "--result-folder",
        help="Where to save the result files. By default it will be saved in a 'result/' folder.",
    ),
    config: Path = typer.Option(
        None,
        "--config",
        help="""Configuration file for column names. File should be a json with keys as your \
            data's column names and values as hilary's required column names.""",
    ),
    override: bool = typer.Option(
        False,
        "--override",
        help="Override existing results.",
    ),
    use_json: bool = typer.Option(
        False,
        "--json/--text",
        help="Print logs as JSON or text.",
    ),
    cdr3only: bool = typer.Option(
        False,
        "--cdr3only",
        help="Use only the cdr3 method.",
    ),
    xy_complete: bool = typer.Option(
        False, "--xy-complete", help="Run xy method on all VJL classes."
    ),
) -> None:
    """Infer lineages from data_path excel file."""
    if result_folder is None:
        result_folder = data_path.parents[0] / Path("hilary_results/")
    result_folder.mkdir(parents=True, exist_ok=True)
    debug_folder = result_folder / Path("debug/")
    debug_folder.mkdir(parents=True, exist_ok=True)

    output_path = result_folder / Path(f"inferred_{data_path.name}")
    if output_path.exists() and not override:
        raise ValueError(
            f"{output_path.as_posix()} already exists, use override parameter to replace the file.",
        )

    if verbose >= 2:  # noqa: PLR2004
        logging_level = logging.DEBUG
    elif verbose == 1:
        logging_level = logging.INFO
    else:
        logging_level = logging.WARNING
    if use_json:
        renderer = structlog.processors.JSONRenderer()
    else:
        renderer = structlog.dev.ConsoleRenderer(sort_keys=False)
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(logging_level),
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            renderer,
        ],
    )
    log = structlog.get_logger()
    log.info(
        "📖 READING DATA ",
        data_path=data_path.as_posix(),
    )
    dataframe = read_input(input_path=data_path, config=config)

    log.debug("Displaying dataframe columns.", columns=dataframe.columns)
    if logging_level == logging.DEBUG:
        input_path = debug_folder / Path(f"input_{data_path.name}")
        log.debug(
            "Saving input dataframe.",
            path=input_path.as_posix(),
        )
        save_dataframe(dataframe=dataframe, save_path=input_path)
    apriori = Apriori(
        dataframe,
        threads=threads,
        precision=precision,
        sensitivity=sensitivity,
        model=model,
        silent=silent,
    )
    if logging_level == logging.DEBUG:
        preprocessed_path = debug_folder / Path(f"preprocessed_input_{data_path.name}")
        log.debug(
            "Saving dataframe used by Hilary after preprocessing.",
            path=preprocessed_path.as_posix(),
        )
        save_dataframe(dataframe=apriori.df, save_path=preprocessed_path)

        classes_path = debug_folder / Path(f"classes_{data_path.name}")
        log.debug(
            "Saving classes used by Hilary after preprocessing.",
            path=classes_path.as_posix(),
        )
        save_dataframe(dataframe=apriori.classes, save_path=classes_path)

    log.info("⏳ COMPUTING HISTOGRAMS ⏳.")
    apriori.get_histograms()

    if logging_level == logging.DEBUG:
        histogram_path = debug_folder / Path(f"histograms_{data_path.name}")
        log.debug("Saving histograms used by Hilary.", path=histogram_path.as_posix())
        save_dataframe(apriori.histograms, histogram_path)

    log.info("⏳ COMPUTING PARAMETERS ⏳.")
    apriori.get_parameters()

    log.info("⏳ COMPUTING THRESHOLDS ⏳.")
    apriori.get_thresholds()

    if logging_level == logging.DEBUG:
        parameters_path = debug_folder / Path(f"parameters_{data_path.name}")
        log.debug(
            "Saving all parameters inferred by Hilary.",
            path=parameters_path.as_posix(),
        )
        save_dataframe(apriori.classes, parameters_path)

    hilary = HILARy(apriori)
    log.info("⏳ COMPUTING PRECISE AND SENSITIVE CLUSTERS ⏳.")
    hilary.compute_prec_sens_clusters()

    if cdr3only:
        log.info("Returning cdr3 method precise clusters.")
        dataframe["family"] = hilary.df["precise_cluster"]

    else:
        log.info("⏳ INFERRING FAMILIES WITH FULL XY METHOD⏳.")
        hilary.infer(xy_complete=xy_complete)

        if logging_level == logging.DEBUG:
            output_path = debug_folder / Path(f"preprocessed_output_{data_path.name}")
            log.debug(
                "Saving dataframe inferred by Hilary.",
                path=output_path.as_posix(),
            )
            save_dataframe(hilary.df, save_path=output_path)
            grouped_by_sensitive = (
                hilary.df.groupby(
                    hilary.group + ["sensitive_cluster", "family"],
                    group_keys=True,
                )[["to_resolve"]]
                .apply(lambda x: x.sum())
                .rename(columns={"to_resolve": "count"})
            )
            grouped_by_sensitive["method"] = grouped_by_sensitive["count"].apply(
                lambda x: "xy" if x > 0 else "cdr3",
            )
            method_summary_path = debug_folder / Path(
                f"method_summary_{data_path.name}"
            )
            log.debug(
                "Saving method distribution summary.",
                path=method_summary_path.as_posix(),
            )
            save_dataframe(
                grouped_by_sensitive["method"], save_path=method_summary_path
            )

        pair_frac = (
            binom(
                hilary.df.query("to_resolve == True")
                .groupby(hilary.group + ["sensitive_cluster"])
                .size(),
                2,
            ).sum()
            / binom(
                hilary.df.groupby(hilary.group + ["sensitive_cluster"]).size(),
                2,
            ).sum()
        )
        log.info(
            "Fraction of pairs that go through xy method.",
            fraction=pair_frac,
        )

        dataframe["family"] = hilary.df["family"]
        output_path = result_folder / Path(f"inferred_{data_path.name}")

        log.info("💾 SAVING RESULTS ", output_path=output_path.as_posix())
        save_dataframe(dataframe=dataframe, save_path=output_path)

        if logging_level == logging.DEBUG and "clone_id" in dataframe.columns:
            precision_full, sensitivity_full = pairwise_evaluation(
                df=dataframe, partition="family"
            )
            hilary.df["clone_id"] = dataframe["clone_id"]
            precision_cdr3, sensitivity_cdr3 = pairwise_evaluation(
                df=hilary.df, partition="precise_cluster"
            )
            log.debug(
                "Evaluating Hilary's performance on ground truth",
                precision_full_method=precision_full,
                sensitivity_full_method=sensitivity_full,
                precision_cdr3=precision_cdr3,
                sensitivity_cdr3=sensitivity_cdr3,
            )


if __name__ == "__main__":
    app()



from __future__ import annotations

from setuptools import find_packages
from setuptools import setup

from hilary import __version__

setup(
    install_requires=[
        "setuptools>=56,<57",
        "numpy>=1.20.0,<2",
        "openpyxl>=3.1,<4",
        "pandas>=2.1,<3",
        "scipy>=1.6,<2",
        "structlog>=22.3.0,<23",
        "textdistance>=4.6,<5",
        "tqdm>=4.66,<5",
        "typer>=0.9,<1",
        "atriegc>=0.0.3,<1.0.0",
        "scipy>1.11,<2",
    ],
    name="hilary",
    version="1.0.20",
    url="https://github.com/statbiophys/HILARy/",
    author="Natanael Spisak, Gabriel Athènes",
    author_email="natanael.spisak@gmail.com, gabriel.athenes@polytechnique.edu",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "infer=hilary.__main__:app",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)
